# -*- coding: utf-8 -*-
"""Set up logging.
"""
#
# Library imports.
#
import logging
import sys
from datetime import datetime
from pathlib import Path

#
# Third-party imports.
#
from flask import request

#
# Global variables.
#
DEFAULT_FILE_LOGLEVEL = logging.INFO
DEFAULT_STDERR_LOGLEVEL = logging.WARNING


class ContextualFilter(logging.Filter):
    """A logging filter with request-based info."""

    def filter(self, record):
        record.utcnow = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S%Z")
        try:
            record.url = request.path
            record.method = request.method
        except RuntimeError:
            record.url = ""
            record.method = ""

        return True


def configure_logging(app):
    """Configure logging to stderr and a log file.

    :param app:
    :return:
    """
    if app.config["DEBUG"]:  # pragma: no cover
        stderr_log_level = logging.DEBUG
    else:
        stderr_log_level = DEFAULT_STDERR_LOGLEVEL
    if app.config["QUIET"]:  # pragma: no cover
        file_log_level = logging.ERROR
    else:
        file_log_level = DEFAULT_FILE_LOGLEVEL
    app.logger.addFilter(ContextualFilter())
    app.logger.setLevel(logging.DEBUG)
    for handler in app.logger.handlers:  # set levels on existing handlers
        handler.setLevel(stderr_log_level)
        handler.setFormatter(
            logging.Formatter(app.config["STDERR_LOG_FORMAT"])
        )
    #
    # Start Sentry monitoring, if SENTRY_DNS is configured.
    #
    if app.config["SENTRY_DSN"] != "":  # pragma: no cover
        from raven.contrib.flask import Sentry

        # import logging
        Sentry(
            app,
            dsn=app.config["SENTRY_DSN"],
            # logging=True,
            # level=logging.ERROR
        )

    #
    # Start log file.
    #
    if app.config["LOGFILE"]:  # start a log file
        logfile_name = app.config["LOGGER_NAME"] + "_errors.log"
        app.config["LOGFILE_NAME"] = logfile_name
        logfile_path = Path(app.config["LOG"]) / logfile_name
        if app.config["DEBUG"]:  # pragma: no cover
            print(f'Logging to file "{str(logfile_path)}".', file=sys.stderr)
        if not logfile_path.parent.is_dir():  # create logs/ dir
            try:  # pragma: no cover
                logfile_path.parent.mkdir(
                    mode=int(app.config["DIR_MODE"], 8), parents=True
                )
            except OSError:  # pragma: no cover
                app.logger.error(
                    'Unable to create logfile directory "%s"',
                    logfile_path.parent,
                )
                raise OSError
        log_handler = logging.handlers.RotatingFileHandler(
            str(logfile_path),
            maxBytes=app.config["LOGFILE_MAXBYTES"],
            backupCount=app.config["LOGFILE_BACKUPCOUNT"],
        )

        log_handler.setLevel(file_log_level)
        werkzeug_logger = logging.getLogger("werkzeug")
        werkzeug_logger.addHandler(log_handler)
        for handler in app.logger.handlers:  # set levels on existing handlers
            handler.setLevel(file_log_level)
            handler.setFormatter(
                logging.Formatter(app.config["STDERR_LOG_FORMAT"])
            )
        app.logger.addHandler(log_handler)
    #
    # Do some logging on startup.
    #
    app.logger.debug('Command line: "%s"', " ".join(sys.argv))
    app.logger.debug(
        "%s version %s", app.config["LOGGER_NAME"], app.config["VERSION"]
    )
    app.logger.debug(
        "Run started at %s", datetime.now().strftime("%Y%m%d-%H%M%S")
    )



# -*- coding: utf-8 -*-
"""Define and handle configuration variables.

Definitions in this file are selected by the MYAPP_CONFIGURATION environmental
variable.  This variable may take on the following values:

* default: starts everything in production environment
* development: starts in debug environment (not secure)
* serverOnly: same as default, but without queues
* treebuilder: starts treebuilder queue only
* aligner: start aligner queue only


These definitions may be overridden via two ways:

1. A python file pointed to by the environmental variable MYAPP_SETTINGS.
2. An environmental variable that starts with MYAPP_.  If its value is
   "True" or "False", then it will be interpreted as a logical value.
   If its value can be parsed as an integer, then it will be.
"""
# Standard-ibrary imports.
import os
import platform
import sys
from getpass import getuser
from importlib import metadata
from socket import getfqdn
from pathlib import Path

# Third-party imports.
import arrow

#
from .common import NAME

#
# Name of this service.
#
SERVICE_NAME = os.getenv("FLASK_APP", __name__.split(".")[0])
SERVICE_ORG = "LegumeFederation"
#
# Definitions that *must* be set in environmental variables.  Trying to
# set these from the config file would be too late, so they are
# not settable by that mechanism.
#
IMMUTABLES = ("ROOT", "VAR", "LOG", "TMP")
PATHVARS = ("ROOT", "VAR", "LOG", "TMP", "DATA", "USERDATA")
PROMETHEUS_SERVICES = (
    "prometheus",
    "alertmanager",
    "node_exporter",
    "push_gateway",
)


def get_path(name, default):
    """Get path from environ, checking absoluteness."""
    varname = SERVICE_NAME.upper() + "_" + name.upper()
    if varname in os.environ:
        path_str = os.environ[varname]
        try:
            Path(path_str).relative_to("/")
        except ValueError:  # relative path, not acceptable
            print(
                f'ERROR--path variable {varname}="{path_str}" not absolute, ignoring'
            )
            path_str = default
    else:
        path_str = default
    return path_str


class BaseConfig(object):
    """Base class for configuration objects.

    Note that only values in uppercase will be stored in the app
    configuration object.
    """

    #
    # Name of this service.
    #
    NAME = SERVICE_NAME
    ENVIRONMENT_DUMP = False
    #
    # Web site associated with this project.
    #
    PROJECT_HOME = "https://github.com/" + SERVICE_ORG + "/" + SERVICE_NAME
    #
    # File path locations.  All of these are immutable except DATA.
    # Since different components run from different locations, these
    # must be absolute.  The immutable ones should be created before
    # runtime.
    #
    ROOT = get_path("ROOT", sys.prefix)
    VAR = get_path("VAR", ROOT + "/var")
    LOG = get_path("LOG", VAR + "/log")
    TMP = get_path("TMP", VAR + "/tmp")
    DATA = get_path("DATA", VAR + "/data/")
    USERDATA = get_path("DATA", VAR + "/userdata/")
    USER_CONFIG_PATH = "~/." + SERVICE_NAME
    #
    # Directory/file permissions.
    #
    PROCESS_UMASK = "0002"
    DIR_MODE = "755"  # Note interaction with process umask
    #
    # The DEBUG parameter has multiple implications:
    #           * access to python debugging via flask
    #           * logging levels set to DEBUG
    #           * configuration variables are printed
    #
    DEBUG = False
    PORT = 58927
    HOST = "localhost"
    #
    # Create a logfile.
    #
    LOGFILE = True
    LOGFILE_NAME = None
    LOGFILE_MAXBYTES = 10000000
    LOGFILE_BACKUPCOUNT = 1
    #
    # Log only errors.
    #
    QUIET = False
    #
    # Test mode, includes propagation of errors.
    #
    TESTING = False
    #
    # User for rc scripts.
    #
    RC_USER = getuser()
    RC_GROUP = ""
    RC_VERBOSE = False
    #
    # Settings file name.
    #
    SETTINGS = SERVICE_NAME + ".conf"
    #
    # Number of threads used in queued commands.  0 = use as many as available.
    #
    THREADS = 0
    #
    # RQ settings.  If "RQ_ASYNC" is False, then no queueing will be done.
    #
    RQ_ASYNC = True
    TREE_QUEUE = "treebuilding"
    ALIGNMENT_QUEUE = "alignment"
    START_QUEUES = []
    RQ_QUEUES = [TREE_QUEUE, ALIGNMENT_QUEUE]
    REDIS_UNIX_SOCKET = True
    RQ_REDIS_PORT = 58929
    RQ_REDIS_HOST = "localhost"
    RQ_SCHEDULER_INTERVAL = 60
    RQ_SCHEDULER_QUEUE = ALIGNMENT_QUEUE
    ALIGNMENT_QUEUE_TIMEOUT = 24 * 60 * 60  # 24 hours, in seconds
    TREE_QUEUE_TIMEOUT = 30 * 24 * 60 * 60  # 30 days, in seconds
    #
    # Definitions for alignment algorithms.
    #
    ALIGNERS = {
        "hmmalign": ["--trim", "--informat", "FASTA"]
    }  # command-line arguments
    HMMALIGN_EXE = "hmmalign"
    #
    # Definitions for tree-building algorithms.
    #
    TREEBUILDERS = {
        "FastTree": {
            "peptide": ["-nopr", "-log", "peptide.log"],
            "DNA": ["-nt", "-gtr", "-log", "nucleotide.log", "-nopr"],
        },
        "RAxML": {
            "peptide": [
                "-b",
                "12345",
                "-p",
                "12345",
                "-N",
                "10",
                "-m",
                "PROTGAMMABLOSUM62",
            ],
            "DNA": ["-d"],
        },
    }
    #
    # Binaries.
    #
    FASTTREE_EXE = "FastTree-" + SERVICE_NAME
    RAXML_EXE = "raxmlHPC"
    #
    # Current run.
    #
    HOSTNAME = getfqdn()
    DATETIME = arrow.now().format("YYYY-MM-DD HH:mm:ss")
    #
    # supervisord defs.
    #
    SUPERVISORD_UNIX_SOCKET = True
    SUPERVISORD_PORT = 58928
    SUPERVISORD_HOST = "localhost"
    SUPERVISORD_USER = SERVICE_NAME
    SUPERVISORD_START_REDIS = True
    SUPERVISORD_START_SERVER = True
    SUPERVISORD_START_REDIS = True
    SUPERVISORD_START_ALIGNMENT = True
    SUPERVISORD_START_TREEBUILDING = True
    SUPERVISORD_START_CRASHMAIL = True
    SUPERVISORD_START_NGINX = True
    SUPERVISORD_START_PROMETHEUS = False
    SUPERVISORD_START_ALERTMANAGER = False
    SUPERVISORD_START_NODE_EXPORTER = False
    SUPERVISORD_START_PUSHGATEWAY = False
    #
    # nginx defs.
    #
    NGINX_SERVER_NAME = "localhost"
    system = platform.system()
    if system == "Linux":
        NGINX_LISTEN_ARGS = "deferred"
        NGINX_EVENTS = "use epoll;"
    elif system.endswith("BSD"):  # pragma: no cover
        NGINX_LISTEN_ARGS = "accept_filter=httpready"
        NGINX_EVENTS = "use kqueue;"
    elif system == "Darwin":  # pragma: no cover
        NGINX_LISTEN_ARGS = ""
        NGINX_EVENTS = "use kqueue;"
    else:  # pragma: no cover
        NGINX_LISTEN_ARGS = ""
        NGINX_EVENTS = ""
    NGINX_UNIX_SOCKET = False
    #
    # gunicorn defs--these will not be used in debugging mode.
    #
    GUNICORN_LOG_LEVEL = "debug"
    GUNICORN_UNIX_SOCKET = True
    #
    # URL defs--these will be used in testing.
    #
    CURL_ARGS = ""
    CURL_URL = HOST + ":" + str(PORT)
    URL = ""
    #
    # crashmail defs.
    #
    CRASHMAIL_EMAIL = getuser()
    CRASHMAIL_EVENTS = "PROCESS_STATE_EXITED"
    #
    # Controls of which processes get started by supervisord.
    # Setting these to empty strings will cause the process to
    # not be started.
    #
    # Monitoring at sentry.io.
    #
    SENTRY_DSN = ""
    #
    # Logging formatter.  Fields that are defined are:
    #    asctime: Time with too much precision
    #    levelname: Severity level.
    #    module:  module name.
    #    lineno: line number.
    #    pathname: File path.
    #    message: The message.
    #    url: Target address (if from a target).
    #    utcnow: Time in UTC.
    #    method: HTTP method.
    #    ip: Real IP address of the requester.
    #
    STDERR_LOG_FORMAT = "%(levelname)s: %(message)s"
    FILE_LOG_FORMAT = "%(levelname)s: %(message)s"


class DevelopmentConfig(BaseConfig):
    """Start internal server, no queues."""

    DEBUG = True
    TESTING = True
    RQ_ASYNC = False
    ENVIRONMENT_DUMP = True
    # Running synchronous--no need to start queues.
    SUPERVISORD_START_REDIS = False
    SUPERVISORD_START_ALIGNMENT = False
    SUPERVISORD_START_TREEBUILDER = False
    # Use debug config settings
    SETTINGS = SERVICE_NAME + "-debug.conf"


class ServerOnlyConfig(BaseConfig):
    """Start server and redis, no queues."""

    SUPERVISORD_START_ALIGNMENT = False
    SUPERVISORD_START_TREEBUILDER = False


class TreebuilderConfig(BaseConfig):
    """Start treebuilder queue only."""

    SUPERVISORD_START_REDIS = False
    SUPERVISORD_START_SERVER = False
    SUPERVISORD_START_ALIGNMENT = False


class AlignerConfig(BaseConfig):
    """Start alignment queue only"""

    SUPERVISORD_START_REDIS = False
    SUPERVISORD_START_SERVER = False
    SUPERVISORD_START_TREEBUILDER = False


#
# Dictionary of configuration levels to be used with
# the _CONFIGURATION environmental variable.
#
config_dict = {
    "default": SERVICE_NAME + ".config.BaseConfig",
    "development": SERVICE_NAME + ".config.DevelopmentConfig",
    "serverOnly": SERVICE_NAME + ".config.ServerOnlyConfig",
    "treebuilder": SERVICE_NAME + ".config.TreebuilderConfig",
    "aligner": SERVICE_NAME + ".config.AlignerConfig",
}


def configure_app(app):
    """Configure the app, getting variables and setting up logging.

    :param app:
    :return:
    """
    config_name = os.getenv(SERVICE_NAME.upper() + "_MODE", "default")
    if config_name not in config_dict:
        print(f'ERROR -- mode "{config_name}" not known.', file=sys.stderr)
        sys.exit(1)
    app.config.from_object(config_dict[config_name])
    app.config["MODE"] = config_name
    #
    # Do overrides from configuration, if it exists.
    #
    app.instance_path = os.getenv(
        SERVICE_NAME.upper() + "_ROOT", app.config["ROOT"]
    )
    pyfile_name = os.getenv(
        SERVICE_NAME.upper() + "_SETTINGS", app.config["SETTINGS"]
    )
    pyfile_path = Path(app.instance_path) / "etc" / pyfile_name
    pyfile_dict = {}
    try:
        with pyfile_path.open(mode="rb") as config_file:
            exec(
                compile(config_file.read(), str(pyfile_path), "exec"),
                pyfile_dict,
            )  # noqa
    except IOError:
        print(f'Unable to load configuration file "{str(pyfile_path)}".')
    for internal_key in ["__doc__", "__builtins__"]:
        if internal_key in pyfile_dict:
            del pyfile_dict[internal_key]
    if "VAR" in pyfile_dict:  # VAR is hierarchical special case
        for subdir in ["tmp", "log", "data", "userdata"]:
            if not subdir.upper() in pyfile_dict:
                pyfile_dict[subdir.upper()] = pyfile_dict["VAR"] + "/" + subdir
    # -*- coding: utf-8 -*-
    for key in pyfile_dict:
        app.config[key] = pyfile_dict[key]
    #
    # Do overrides from environmental variables.
    #
    for my_envvar, envvar in [
        (i, i[6:])
        for i in sorted(os.environ)
        if i.startswith(SERVICE_NAME.upper() + "_")
    ]:
        value = os.environ[my_envvar]
        if value == "True":
            value = True
        elif value == "False":
            value = False
        else:
            try:
                value = int(value)
            except ValueError:
                pass
        if envvar not in PATHVARS:  # paths already configured from envvars
            app.config[envvar] = value
    #
    # Set version and platform (output only, not configurable).
    #
    app.config["VERSION"] = metadata.version(NAME)
    app.config["PLATFORM"] = platform.system()
    #
    # Prometheus services.
    #
    for service in PROMETHEUS_SERVICES:
        root_path = Path(app.config["ROOT"])
        try:
            app.config[service.upper() + "_DIR"] = str(
                [p for p in root_path.glob(service + "*") if p.is_dir()][0]
            )
        except IndexError:
            app.config[service.upper() + "_DIR"] = None
    #
    # Set redis socket type.
    #
    if app.config["REDIS_UNIX_SOCKET"]:
        app.config["RQ_REDIS_PORT"] = 0
        app.config["RQ_REDIS_HOST"] = "localhost"
        app.config["RQ_REDIS_URL"] = (
            "unix://@'" + app.config["VAR"] + "/run/redis.sock?db=0"
        )
        app.config[
            "RQ_UNIXSOCKET"
        ] = f"unixsocket {app.config['VAR']}/run/redis.sock"
    else:
        app.config["RQ_REDIS_URL"] = (
            "redis://"
            + app.config["RQ_REDIS_HOST"]
            + ":"
            + str(app.config["RQ_REDIS_PORT"])
            + "/0"
        )
        app.config["RQ_UNIXSOCKET"] = ""
    app.config["REDIS_HOST"] = app.config["RQ_REDIS_HOST"]
    app.config["REDIS_PORT"] = app.config["RQ_REDIS_PORT"]
    app.config["REDIS_URL"] = app.config["RQ_REDIS_URL"]
    #
    # Supervisord socket type.
    #
    if app.config["SUPERVISORD_UNIX_SOCKET"]:
        app.config["SUPERVISORD_SERVERURL"] = (
            "unix://%(ENV_"
            + SERVICE_NAME.upper()
            + "_VAR)s/run/supervisord.sock"
        )
    else:
        app.config["SUPERVISORD_SERVERURL"] = (
            "http://"
            + app.config["SUPERVISORD_HOST"]
            + ":"
            + str(app.config["SUPERVISORD_PORT"])
        )
    #
    # Gunicorn socket type.
    #
    if not app.config["SUPERVISORD_START_NGINX"]:
        app.config["NGINX_URL"] = "noURL"
        if app.config["GUNICORN_UNIX_SOCKET"]:
            app.config["URL"] = (
                "unix://" + app.config["VAR"] + "/run/gunicorn.sock"
            )
            app.config["GUNICORN_URL"] = (
                "unix://%(ENV_"
                + SERVICE_NAME.upper()
                + "_VAR)s/run/gunicorn.sock"
            )
            app.config["CURL_ARGS"] = (
                "--unix-socket " + app.config["VAR"] + "/run/gunicorn.sock"
            )
            app.config["CURL_URL"] = "http://localhost"
        else:
            app.config["URL"] = (
                "http://" + app.config["HOST"] + ":" + str(app.config["PORT"])
            )
            app.config["GUNICORN_URL"] = (
                app.config["HOST"] + ":" + str(app.config["PORT"])
            )
            app.config["CURL_URL"] = (
                app.config["HOST"] + ":" + str(app.config["PORT"])
            )
    else:  # serve gunicorn to nginx through a socket
        app.config["GUNICORN_UNIX_SOCKET"] = True
        app.config["GUNICORN_URL"] = (
            "unix://%(ENV_" + SERVICE_NAME.upper() + "_VAR)s/run/gunicorn.sock"
        )
        if app.config["NGINX_UNIX_SOCKET"]:
            app.config["URL"] = (
                "unix://" + app.config["VAR"] + "/run/nginx.sock"
            )
            app.config["NGINX_URL"] = (
                "unix://" + app.config["VAR"] + "/run/nginx.sock"
            )
            app.config["CURL_ARGS"] = (
                "--unix-socket " + app.config["VAR"] + "/run/nginx.sock"
            )
            app.config["CURL_URL"] = (
                "http://" + app.config["NGINX_SERVER_NAME"]
            )
        else:
            app.config["URL"] = (
                "http://" + app.config["HOST"] + ":" + str(app.config["PORT"])
            )
            app.config["NGINX_URL"] = (
                app.config["HOST"] + ":" + str(app.config["PORT"])
            )
            app.config["CURL_URL"] = (
                app.config["HOST"] + ":" + str(app.config["PORT"])
            )
    #
    # Set queues to be started.
    #
    for queue in app.config["RQ_QUEUES"]:
        if app.config["SUPERVISORD_START_" + queue.upper()]:
            app.config["START_QUEUES"].append(queue)


def print_config_var(app, var, config_file_obj):
    """Print configuration variable with type and provenance.

    :param var:
    :param obj:
    :return:
    """
    if __name__.upper() + "_" + var in os.environ:
        source = " # <- from environment"
    elif var in config_file_obj.__dict__:
        source = " # <- from config file"
    else:
        source = ""
    val = app.config[var]
    if isinstance(val, str):
        quote = '"'
    else:
        quote = ""
    print(
        f"  {var} type({type(val).__name__}) =  {quote}{val}{quote} {source}"
    )



# -*- coding: utf-8 -*-
"""Command-line interpreter functions.

These functions extend the Flask CLI.

"""
#
# Standard library imports.
#
import json
import os
import pkg_resources
import pkgutil
import sys
import types
from distutils.util import strtobool
from pydoc import locate
from pathlib import Path  # python 3.4 or later

#
#
# third-party imports.
import click
from flask import current_app
from flask.cli import FlaskGroup
from jinja2 import Environment, PackageLoader

#
# Local imports.
#
from .logging import configure_logging
from .filesystem import init_filesystem
from .config_file import create_config_file, write_kv_to_config_file
from .config import print_config_var

#
# Global variables.
#
AUTHOR = "Joel Berendzen"
EMAIL = "JBerendzen@salud.unm.edu"
COPYRIGHT = """Copyright (C) 2021, The University of New Mexico.
All rights reserved.
"""
#
# CLI entry point.
#


@click.group(cls=FlaskGroup, epilog=AUTHOR + " <" + EMAIL + ">. " + COPYRIGHT)
def cli():
    pass


@cli.command()
def run():  # pragma: no cover
    """Run a server directly."""
    from .logging import configure_logging

    print(
        "Direct start, use of gunicorn is recommended for production.",
        file=sys.stderr,
    )
    port = current_app.config["PORT"]
    host = current_app.config["HOST"]
    debug = current_app.config["DEBUG"]
    init_filesystem(current_app)
    configure_logging(current_app)
    current_app.run(host=host, port=port, debug=debug)


@cli.command()
@click.option(
    "--vartype",
    help="Type of variable, if not previously defined.",
    default=None,
)
@click.option("--verbose/--no-verbose", help="Verbose provenance.")
@click.option(
    "--delete/--no-delete", help="Deletes config file, arguments ignored."
)
@click.argument("var", required=False)
@click.argument("value", required=False)
def config(var, value, vartype, verbose, delete):
    """Gets, sets, or deletes config variables."""
    config_file_path = (
        Path(current_app.config["ROOT"])
        / "etc"
        / current_app.config["SETTINGS"]
    )
    if delete:
        if config_file_path.exists():
            print(f"Deleting config file {str(config_file_path)}.")
            config_file_path.unlink()
            create_config_file(config_file_path)
            sys.exit(0)
        else:
            print(
                f"ERROR--config file {str(config_file_path)} does not exist."
            )
            sys.exit(1)
    if value is None:  # No value specified, this is a get.
        config_file_obj = types.ModuleType("config")  # noqa
        if config_file_path.exists():
            config_file_status = "exists"
            config_file_obj.__file__ = str(config_file_path)
            try:
                with config_file_path.open(mode="rb") as config_file:
                    exec(
                        compile(
                            config_file.read(), str(config_file_path), "exec"
                        ),
                        config_file_obj.__dict__,
                    )
            except IOError as e:
                e.strerror = (
                    f"Unable to load configuration file ({e.strerror})"
                )
                raise
        else:
            config_file_status = "does not exist"
            config_file_obj.__file__ = None
        if var is None:  # No variable specified, list them all.
            print(
                "The instance-specific config file at %s %s."
                % (str(config_file_path), config_file_status)
            )
            print(
                "Listing all %d defined configuration variables:"
                % (len(current_app.config))
            )
            for key in sorted(current_app.config):
                print_config_var(current_app, key, config_file_obj)
            return
        else:
            var = var.upper()
            if var.startswith(__name__.upper() + "_"):
                var = var[len(__name__) + 1 :]
            if var in current_app.config:
                if verbose:
                    print_config_var(current_app, var, config_file_obj)
                else:
                    print(current_app.config[var])
                return
            else:
                print(
                    f'"{var}" not found in configuration variables.',
                    file=sys.stderr,
                )
                sys.exit(1)
    else:  # Must be setting.
        var = var.upper()
        if var.startswith(__name__.upper() + "_"):
            var = var[len(__name__) + 1 :]
        old_value = None
        if (
            var in current_app.config
            and vartype is None
            and not current_app.config[var] is None
        ):  # type defaults to current type
            old_value = current_app.config[var]
            value_type = type(old_value)
        else:  # get type from command line, or str if not specified
            if vartype is None:
                vartype = "str"
            value_type = locate(vartype)
        if value_type == bool:
            value = bool(strtobool(value))
        elif value_type == str:
            pass
        else:  # load through JSON to handle dict and list types
            try:
                jsonobj = json.loads(value)
            except json.decoder.JSONDecodeError:
                print(
                    f'ERROR--Unparseable string "{value}". Did you use quotes?',
                    file=sys.stderr,
                )
                sys.exit(1)
            try:
                value = value_type(jsonobj)
            except TypeError:
                print(
                    'ERROR--Unable to convert "%s" of type %s to type %s.'
                    % (value, type(jsonobj).__name__, value_type.__name__),
                    file=sys.stderr,
                )
                sys.exit(1)
        #
        # Write key/value pair to config file.
        #
        create_config_file(config_file_path)
        write_kv_to_config_file(
            config_file_path, var, value, value_type, old_value
        )


@cli.command()
def test_logging():
    """Test logging at the different levels."""
    configure_logging(current_app)
    current_app.logger.debug("Debug message.")
    current_app.logger.info("Info message.")
    current_app.logger.warning("Warning message.")
    current_app.logger.error("Error message.")


def walk_package(root):
    """Walk through a package_resource.

    :type module_name: basestring
    :param module_name: module to search in
    :type dirname: basestring
    :param dirname: base directory
    """
    dirs = []
    files = []
    for name in pkg_resources.resource_listdir(__name__, root):
        fullname = root + "/" + name
        if pkg_resources.resource_isdir(__name__, fullname):
            dirs.append(fullname)
        else:
            files.append(name)
    for new_path in dirs:
        yield from walk_package(new_path)
    yield root, dirs, files


def copy_files(pkg_subdir, out_head, force, notemplate_exts=None):
    """Copy files from package, with templating.

    :param pkg_subdir:
    :param out_head:
    :param force:
    :param notemplate_exts:
    :return:
    """
    for root, dirs, files in walk_package(pkg_subdir):
        del dirs
        split_dir = os.path.split(root)
        if split_dir[0] == "":
            out_subdir = ""
        else:
            out_subdir = "/".join(list(split_dir)[1:])
        out_path = out_head / out_subdir
        if not out_path.exists() and len(files) > 0:
            print(f'Creating "{str(out_path)}" directory')
            out_path.mkdir(
                mode=int(current_app.config["DIR_MODE"], 8), parents=True
            )
        #
        # Initialize Jinja2 template engine on this directory.
        #
        template_env = Environment(
            loader=PackageLoader('rayon', root),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        for filename in files:
            try:
                ext = os.path.splitext(filename)[1].lstrip(".")
            except IndexError:
                ext = ""
            if notemplate_exts is not None and ext in notemplate_exts:
                templated = "directly"
                data_string = pkgutil.get_data(
                    __name__, root + "/" + filename
                ).decode("UTF-8")
            else:
                templated = "from template"
                template = template_env.get_template(filename)
                data_string = template.render(current_app.config)
            outfilename = filename.replace(
                "server", current_app.config["NAME"]
            )
            file_path = out_path / outfilename
            if file_path.exists() and not force:
                print(
                    f"ERROR -- File {str(file_path)} already exists."
                    + "  Use --force to overwrite."
                )
                sys.exit(1)
            elif file_path.exists() and force:
                operation = "Overwriting"
            else:
                operation = "Creating"
            with file_path.open(mode="wt") as fh:
                print(f'{operation} file "{str(file_path)}" {templated}.')
                fh.write(data_string)
            if (
                filename.endswith(".sh")
                or filename == current_app.config["NAME"]
            ):
                file_path.chmod(0o755)


@cli.command()
@click.option(
    "--force/--no-force",
    help="Force overwrites of existing files",
    default=False,
)
@click.option("--init/--no-init", help="Initialize filesystem", default=True)
@click.option(
    "--var/--no-var", help="Create files in var directory", default=True
)
def create_instance(force, init, var):
    """Configures instance files."""
    copy_files("etc", Path(current_app.config["ROOT"]) / "etc", force)
    if var:
        copy_files("var", Path(current_app.config["VAR"]), force)
    if init:
        init_filesystem(current_app)


@cli.command()
@click.option(
    "--force/--no-force",
    help="Force overwrites of existing files",
    default=False,
)
def set_htpasswd(force):
    """Sets the site password to SECRET_KEY."""
    htpasswd_file = current_app.config["ROOT"] + "/etc/nginx/htpasswd"
    htpasswd_path = Path(htpasswd_file)
    user = current_app.config["NAME"]
    secret_key = current_app.config["SECRET_KEY"]
    print(f"Setting password for user {user} to {secret_key}. ")
    if not htpasswd_path.exists():
        print("Creating htpasswd file.")
        htpasswd_path.touch()
    with htpasswd.Basic(htpasswd_file) as userdb:
        if len(secret_key) == 0:
            print("ERROR--must set SECRET_KEY first.")
            sys.exit(1)
        try:
            userdb.add(user, secret_key)
        except htpasswd.basic.UserExists:
            if force:
                print(f"Updating site password for existing user {user}.")
                userdb.change_password(user, secret_key)
            else:
                print(
                    "ERROR--user %s already exists in htpasswd, use --force."
                )
                sys.exit(1)


@cli.command()
@click.option(
    "--force/--no-force",
    help="Force overwrites of existing files",
    default=False,
)
@click.option(
    "--configonly/--no-configonly",
    help="Only create config file",
    default=False,
)
def create_test_files(force, configonly):
    """Create test files."""
    if not configonly:
        copy_files("test", Path("."), force, notemplate_exts=["sh"])
    copy_files(
        "user_conf",
        Path(os.path.expanduser(current_app.config["USER_CONFIG_PATH"])),
        force,
    )



# -*- coding: utf-8 -*-
"""Set configuration file from setup.py or server"""
import string
import sys
from datetime import datetime
from secrets import choice


DEFAULT_PASSWORD_LENGTH = 12
DEFAULT_GROUPING = 4
CONFIG_EXT = ".conf"
CONFIG_FILE_HEADER = """# -*- coding: utf-8 -*-
'''Overrides of default configurations.

This file is sourced after default configs but before environmental variables.

You may hand-edit this file, but it may be deleted with the config --delete
command.  Moreover, further sets will append and possibly supercede hand-edited
values.  

Note that configuration variables are all-caps.

Types are derived from python typing rules.
'''"""  # noqa


def generate_random_password(
    length=DEFAULT_PASSWORD_LENGTH, grouping=DEFAULT_GROUPING
):
    """Generate a password from an alphabet"""
    alphabet = string.ascii_lowercase + string.digits
    nchars = 0
    secret_key = ""
    while nchars < length:
        secret_key += choice(alphabet)  # noqa
        nchars += 1
        if grouping and nchars < length and not nchars % grouping:
            secret_key += "-"
    return secret_key


def write_kv_to_config_file(file_path, key, value, valtype, previous_value):
    if valtype is str:
        quote = '"'
    else:
        quote = ""
    print(
        '%s was %s%s%s, now set to %s%s%s (type %s) \n in config file "%s".'
        % (
            key,
            quote,
            previous_value,
            quote,
            quote,
            value,
            quote,
            type(value).__name__,
            str(file_path),
        )
    )
    with file_path.open(mode="a") as config_fh:
        isodate = datetime.now().isoformat()[:-7]
        print(
            f"{key} = {quote}{value}{quote} # set at {isodate}",
            file=config_fh,
        )  # noqa


def create_config_file(file_path):
    """Initializes config file with secret key."""
    dir_path = file_path.parent
    if not dir_path.is_dir():  # pragma: no cover
        print(f'Creating application etc/ directory "{str(dir_path)}".')
        dir_path.mkdir(mode=0o775, parents=True)
    if not file_path.exists():
        with file_path.open(mode="w") as config_fh:
            print(f'Creating instance config file at "{str(file_path)}".')
            print(CONFIG_FILE_HEADER, file=config_fh)
        #
        # Set SECRET_KEY to a random string.
        #
        write_kv_to_config_file(
            file_path, "SECRET_KEY", generate_random_password(), str, ""
        )


if __name__ == "__main__":  # for development and testing purposes only
    from pathlib import Path  # python > 3.4

    create_config_file(Path(sys.prefix) / "settings.conf")
